# import random
# import numpy as np

# from .ValueFunction import *
# from .ExperimentalValueFunction import *
# from .ICF import *
# from .LinearUCB import *
# from .LinearThompsonSampling import *
# from .LinearEGreedy import *
# from .GLM_UCB import *
# from .MostPopular import *
# from .Random import *
# from .Entropy import *
# from .LogPopEnt import *
# from .ThompsonSampling import *
# from .EGreedy import *
# from .UCB import *
# from .LinUCB import *
# from .UCBLearner import *
# from .MostRepresentative import *
# from .LinEGreedy import *
# from .OurMethod1 import *
# from .Entropy0 import *
# from .HELF import *
# from .PopPlusEnt import *
# from .EMostPopular import *
# from .DistinctPopular import *
# from .PPELPE import *
# from .kNNBandit import *
# from .COFIBA import *
# from .OurMethod2 import *
# from .PTS import *
# from .MFValueFunction import *
# from .NICF import NICF
# from .OurMethod3 import OurMethod3
# from .OurMethod4 import OurMethod4
# from .OurMethod5 import OurMethod5
# from .OurMethodBalanced import OurMethodBalanced
# from .OurMethodInit import *
# from .LinearUCB1 import *
# from .LinUCBvar import *
# from .OurMethod2var import *
# from .LinearUCBInit import *
# from .LinearEGreedyInit import *
# from .OurMethod1Mod import *
# from .OurMethod2PCA import *
# from .LinUCBPCA import *
# from .ICTR import *
# from .GenericThompsonSampling import *
